@@ -1,18 +1,12 @@
 package CGI::Application::Dispatch;
 use strict;
 use warnings;
-use Carp qw(confess carp);
+use Carp qw(carp cluck);
 use Exception::Class::TryCatch qw(catch);
 
-our $VERSION = '2.00_06';
+our $VERSION = '2.10_01';
 our $DEBUG = 0;
 
-# Used for error handling
-
-# a cache
-our %URL_DISPATCH_CACHE = ();
-our $URL_DISPATCH_CACHE_MAX = 1000;
-
 BEGIN {
     use constant IS_MODPERL  => exists( $ENV{MOD_PERL} );
     use constant IS_MODPERL2 => (
@@ -39,17 +33,17 @@ use Exception::Class (
     'CGI::Application::Dispatch::ERROR' => {
         isa         => 'CGI::Application::Dispatch::Exception',
         alias       => 'throw_error',
-        description => (IS_MODPERL() ? SERVER_ERROR() : 500 ),
+        description => 500,
     },
     'CGI::Application::Dispatch::NOT_FOUND' => {
         isa         => 'CGI::Application::Dispatch::Exception',
         alias       => 'throw_not_found',
-        description => ( IS_MODPERL() ? NOT_FOUND() : 404 ),
+        description => 404,
     },
     'CGI::Application::Dispatch::BAD_REQUEST' => {
         isa         => 'CGI::Application::Dispatch::Exception',
         alias       => 'throw_bad_request',
-        description => ( IS_MODPERL2() ? HTTP_BAD_REQUEST() : IS_MODPERL ? BAD_REQUEST : 400 ),
+        description => 400,
     },
 );
 
@@ -87,7 +81,7 @@ Under normal cgi
             prefix  => 'MyApp',
             table   => [
                 ''                => { app => 'Welcome', rm => 'start' },
-                :app/:rm'         => { },
+                ':app/:rm'        => { },
                 'admin/:app/:rm'  => { prefix   => 'MyApp::Admin' },
             ],
         };
@@ -191,6 +185,20 @@ exist.
 If this value is not present, then Dispatch will return a NOT FOUND error either to the 
 browser (under CGI) or to Apache (under mod_perl).
 
+=item auto_rest
+
+This tells Dispatch that you are using REST by default and that you care about which HTTP method
+is being used. Dispatch will append the HTTP method name (upper case by default) to
+the run mode that is determined after finding the appropriate dispatch rule. So a GET request
+that translates into C<MyApp::Module->foo> will become C<MyApp::Module->foo_GET>.
+
+This can be overridden on a per-rule basis in a custom dispatch table.
+
+=item auto_rest_lc
+
+In combinaion with L<auto_rest> this tells Dispatch that you prefer lower cased HTTP method names.
+So instead of C<foo_POST> and C<foo_GET> you'll have C<foo_post> and C<foo_get>.
+
 =back
 
 =cut
@@ -198,29 +206,73 @@ browser (under CGI) or to Apache (under mod_perl).
 sub dispatch {
     my ($self, %args) = @_;
 
-    %args = ( %{ $self->dispatch_args(\%args) }, %args);
+    # merge dispatch_args() and %args with %args taking precendence
+    my $dispatch_args = $self->dispatch_args(\%args);
+    foreach my $arg (keys %$dispatch_args) {
+        # args_to_new should be merged
+        if( $arg eq 'args_to_new' ) {
+            $args{args_to_new} ||= {};
 
-    $DEBUG = 1 if $args{debug};
+            # merge the PARAMS hash
+            if( $dispatch_args->{args_to_new}->{PARAMS} ) {
+                # merge the hashes
+                $args{args_to_new}->{PARAMS} = {
+                    %{$dispatch_args->{args_to_new}->{PARAMS}},
+                    %{ $args{args_to_new}->{PARAMS} || {} },
+                }
+            }
 
-    # Immediatey, try a cached result.
-    if ( my $final_args = $self->_url_cache() ) {
-        if( $DEBUG ) {
-            require Data::Dumper;
-            warn "[Dispatch] - Found cached version of URL '$ENV{REQUEST_URI}'. "
-                . "Using the following args: " . Data::Dumper::Dumper($final_args);
+            # combine any TMPL_PATHs
+            if( $dispatch_args->{args_to_new}->{TMPL_PATH} ) {
+                # make sure the orginial is an array ref
+                if( $args{args_to_new}->{TMPL_PATH} ) {
+                    if( ! ref $args{args_to_new}->{TMPL_PATH} ) {
+                        $args{args_to_new}->{TMPL_PATH} = [
+                            $args{args_to_new}->{TMPL_PATH}
+                        ]; 
+                    }
+                } else {
+                    $args{args_to_new}->{TMPL_PATH} = [];
+                }
+
+                # now add the rest to the end
+                if( ref $dispatch_args->{args_to_new}->{TMPL_PATH} ) {
+                    push(
+                        @{$args{args_to_new}->{TMPL_PATH}},
+                        @{$dispatch_args->{args_to_new}->{TMPL_PATH}},
+                    );
+                } else {
+                    push(
+                        @{$args{args_to_new}->{TMPL_PATH}},
+                        $dispatch_args->{args_to_new}->{TMPL_PATH},
+                    );
+                }
+            }
+
+            # now merge the args_to_new hashes
+            $args{args_to_new} = {
+                %{$dispatch_args->{args_to_new}},
+                %{$args{args_to_new}},
+            };
+        } else {
+            # anything else should override
+            $args{$arg} = $dispatch_args->{$arg} unless exists $args{$arg};
         }
-        return $self->_run_app(@$final_args);
     }
 
+    $DEBUG = $args{debug} ? 1 : 0;
+
     # check for extra args (for backwards compatibility)
     foreach (keys %args) {
         next if(
-            $_ eq 'prefix'      or
-            $_ eq 'default'     or
-            $_ eq 'debug'       or
-            $_ eq 'rm'          or
-            $_ eq 'args_to_new' or
-            $_ eq 'table'       or
+            $_ eq 'prefix'       or
+            $_ eq 'default'      or
+            $_ eq 'debug'        or
+            $_ eq 'rm'           or
+            $_ eq 'args_to_new'  or
+            $_ eq 'table'        or
+            $_ eq 'auto_rest'    or
+            $_ eq 'auto_rest_lc' or
             $_ eq 'not_found'
         );
         carp "Passing extra args ('$_') to dispatch() is deprecated! Please use 'args_to_new'";
@@ -242,11 +294,12 @@ sub dispatch {
     my $named_args;
     eval {
         $named_args = $self->_parse_path( $path_info, $args{table} )
-		    or throw_not_found("Could not parse PATH_INFO");
+		    or throw_not_found(
+                "Resource not found " . ($ENV{REQUEST_URI} ? "for '$ENV{REQUEST_URI}'" : '')
+            );
     }; 
-    my $e;
-    catch( $e, [ 'CGI::Application::Dispatch::Exception' ])
-        and return $self->http_error($e, $args{not_found});
+    my $e = catch();
+    return $self->http_error($e, $args{not_found}) if( $e );
 
     if ($DEBUG) {
         require Data::Dumper;
@@ -265,6 +318,14 @@ sub dispatch {
         ( $module, $local_prefix, $rm, $local_args_to_new ) =
           delete @{$named_args}{qw(app prefix rm args_to_new)};
 
+        # If another name for dispatch_url_remainder has been set move
+        # the value to the requested name
+        if($$named_args{'*'}) {
+          $$named_args{$$named_args{'*'}} = $$named_args{'dispatch_url_remainder'};
+          delete $$named_args{'*'};
+          delete $$named_args{'dispatch_url_remainder'};
+        }
+
         $module or throw_error("App not defined");
         $module = $self->translate_module_name($module);
 
@@ -276,22 +337,21 @@ sub dispatch {
         # add the rest of the named_args to PARAMS
         @{ $local_args_to_new->{PARAMS} }{ keys %$named_args } = values %$named_args;
 
+        my $auto_rest = defined $named_args->{auto_rest} ? $named_args->{auto_rest} : $args{auto_rest};
+        if( $auto_rest ) {
+            my $method_lc = defined $named_args->{auto_rest_lc} ? $named_args->{auto_rest_lc} : $args{auto_rest_lc};
+            my $http_method = $self->_http_method;
+            $http_method = lc $http_method if $method_lc;
+            $rm .= "_$http_method"; 
+        }
+
         # load and run the module
         @final_dispatch_args = ($module, $rm, $local_args_to_new);
         $self->require_module($module);
         $output = $self->_run_app($module,$rm,$local_args_to_new);
     };
     $e = catch();
-    if( $e ) {
-        if( $e->isa('CGI::Application::Dispatch::Exception') ) {
-            return $self->http_error($e, $args{not_found});
-        } else {
-            confess $e;
-        }
-    }
-
-    # Cache this URL - dispatch map for later use.
-    $self->_url_cache(\@final_dispatch_args);
+    return $self->http_error($e, $args{not_found}) if( $e );
 
     return $output;
 }
@@ -360,11 +420,16 @@ sub http_error {
         $redirect = 1;
         warn "[Dispatch] Redirection for NOT FOUND to $url" if( $DEBUG );
     }
-    my $errno = $e->description;
+    my $errno;
+    if( $e->isa('CGI::Application::Dispatch::Exception') ) {
+        $errno = $e->description;
+    } else {
+        $errno = 500;
+    }
 
     # if we're under mod_perl
     if (IS_MODPERL) {
-        my $r = IS_MODPERL2 ? Apache2::RequestUtil->request : Apache->request;
+        my $r = $self->_r;
         # if we just want to redirect
         if( $redirect ) {
             $r->status( $errno );
@@ -406,7 +471,6 @@ sub http_error {
 }
 
 
-
 # protected method - designed to be used by sub classes, not by end users
 sub _parse_path {
     my ( $self, $path, $table ) = @_;
@@ -419,17 +483,28 @@ sub _parse_path {
         return;
     }
 
+    # look at each rule and stop when we get a match
     for ( my $i = 0 ; $i < scalar(@$table) ; $i += 2 ) {
 
-        # translate the rule into a regular expression, but remember where the named args are
         my $rule = $table->[$i];
 
+        # are we trying to dispatch based on HTTP_METHOD?
+        my $http_method_regex = qr/\[([^\]]+)\]$/;
+        if( $rule =~ /$http_method_regex/ ) {
+            my $http_method = $1;
+            # go ahead to the next rule
+            next unless lc($1) eq lc($self->_http_method);
+            # remove the method portion from the rule
+            $rule =~ s/$http_method_regex//;
+        }
+
         # make sure they start and end with a '/' to match how PATH_INFO is formatted
         $rule = "/$rule" unless ( index( $rule, '/' ) == 0 );
         $rule = "$rule/" if ( substr( $rule, -1 ) ne '/' );
 
         my @names = ();
 
+        # translate the rule into a regular expression, but remember where the named args are
         # '/:foo' will become '/([^\/]*)'
         # and
         # '/:bar?' will become '/?([^\/]*)?'
@@ -443,6 +518,13 @@ sub _parse_path {
             $1 . ($4 ? '?([^/]*)?' : '([^/]*)')
         }gxe;
 
+        # '/*/' will become '/(.*)/$' the end / is added to the end of
+        # both $rule and $path elsewhere
+        if($rule =~ m{/\*/$}) {
+          $rule =~ s{/\*/$}{/(.*)/\$};
+          push(@names,'dispatch_url_remainder');
+        } 
+
         warn "[Dispatch] Trying to match '${path}' against rule '$table->[$i]' using regex '${rule}'\n"
 			if $DEBUG;
 
@@ -461,31 +543,9 @@ sub _parse_path {
     return;
 }
 
-# protected method - designed to be used by sub classes, not by end users
-sub _url_cache {
-    my ( $self, $value ) = @_;
-    my $key;
+sub _http_method { IS_MODPERL ? shift->_r->method : $ENV{HTTP_REQUEST_METHOD}; }
 
-    # may be PATH_TRANSLATED only to use here?
-    foreach my $env qw(SERVER_NAME SCRIPT_NAME PATH_INFO) {
-        $key .= ":" . ($ENV{$env} || '');
-    }
-
-    if ( defined($value) ) {
-        # shrink cache, if it's too large
-        delete(
-            @URL_DISPATCH_CACHE{
-                ( keys %URL_DISPATCH_CACHE )
-                  [ 0 .. int( $URL_DISPATCH_CACHE_MAX / 3 ) ]
-              }
-          )
-          if ( keys %URL_DISPATCH_CACHE > $URL_DISPATCH_CACHE_MAX );
-
-        $URL_DISPATCH_CACHE{$key} = $value;
-    }
-
-    return $URL_DISPATCH_CACHE{$key};
-}
+sub _r { IS_MODPERL2 ? Apache2::RequestUtil->request : Apache->request; }
 
 sub _run_app {
     my ( $self, $module, $rm, $args ) = @_;
@@ -509,7 +569,7 @@ sub _run_app {
     if ($@) {
         # catch invalid run-mode stuff
         if ( $@ =~ /No such run mode/ ) {
-            throw_not_found('RM not found') 
+            throw_not_found("RM '$rm' not found") 
         # otherwise, just pass it up the chain
         } else {
             die $@;
@@ -528,7 +588,7 @@ hash of new()
     <Location /app>
         SetHandler perl-script
         PerlHandler CGI::Application::Dispatch
-        LerlSetVar  CGIAPP_DISPATCH_PREFIX  MyApp
+        PerlSetVar  CGIAPP_DISPATCH_PREFIX  MyApp
         PerlSetVar  CGIAPP_DISPATCH_DEFAULT /module_name
     </Location>
 
@@ -606,7 +666,15 @@ sub handler : method {
 
     $self->dispatch(%args);
 
-    return $r->status || OK();
+    if( $r->status == 404 ) {
+        return NOT_FOUND();
+    } elsif( $r->status == 500 ) {
+        return SERVER_ERROR();
+    } elsif( $r->status == 400 ) {
+        return BAD_REQUEST();
+    } else {
+        return OK();
+    }
 }
 
 =head2 dispatch_args
@@ -737,7 +805,7 @@ Sometimes it's easiest to explain with an example, so here you go:
         TMPL_PATH => 'myapp/templates'
     },
     table       => [
-        ''                         => { app => 'Blog', rm => 'recent'}
+        ''                         => { app => 'Blog', rm => 'recent'},
         'posts/:category'          => { app => 'Blog', rm => 'posts' },
         ':app/:rm/:id'             => { app => 'Blog' },
         'date/:year/:month?/:day?' => {
@@ -829,6 +897,29 @@ the application, if they existed in the URL.
         ...
     }
 
+=item wildcard
+
+The wildcard token "*" allows for partial matches. The token MUST appear at the end of the
+rule. 
+
+  'posts/list/*'
+
+By default, the C<dispatch_url_remainder> param is set to the remainder of the URL
+matched by the *. The name of the param can be changed by setting "*" argument in the 
+L<ARG LIST>.
+
+  'posts/list/*' => { '*' => 'post_list_filter' }
+
+=item method
+
+You can also dispatch based on HTTP method. This is similar to using L<auto_rest> but
+offers more fine grained control. You include the method (case insensitive) at the end of 
+the rule and enclose it in square brackets.
+
+  ':app/news[post]'   => { rm => 'add_news'    },
+  ':app/news[get]'    => { rm => 'news'        },
+  ':app/news[delete]' => { rm => 'delete_news' },
+
 =back
 
 The main reason that we don't use regular expressions for dispatch rules is that regular
@@ -943,7 +1034,7 @@ Perhaps it's specific to using C<RewriteBase>. Refactorings welcome!
   RewriteCond %{REQUEST_FILENAME} !-d
 
   # Otherwise, pass everything through to the dispatcher
-  RewriteRule ^(.*)$ /cgi-bin/dispatch.cgi$1 [L,QSA]
+  RewriteRule ^(.*)$ /cgi-bin/dispatch.cgi/$1 [L,QSA]
 
 =head2 More complex rewrite: dispatching "/" and multiple developers
 
@@ -982,7 +1073,7 @@ user blocks.
     RewriteCond %{REQUEST_FILENAME} !-d
 
      # Otherwise, pass everything through to the dispatcher
-    RewriteRule ^(.*)$ /~mark/cgi-bin/dispatch.cgi$1 [L,QSA]
+    RewriteRule ^(.*)$ /~mark/cgi-bin/dispatch.cgi/$1 [L,QSA]
 
     # These examples may also be helpful, but are unrelated to dispatching.
     SetEnv DEVMODE mark
@@ -997,33 +1088,14 @@ While Dispatch tries to be flexible, it won't be able to do everything that peop
 we've made it flexible enough so that if it doesn't do I<The Right Thing> you can easily subclass
 it.
 
-=head2 PROTECTED METHODS
+=cut
 
-The following methods are intended to be overridden by subclasses if necessary. They are not
-part of the public API since end users will never touch them. However, to ensure that your
-subclass of Dispatch does not break with a new release, they are documented here and are considered
-to be part of the API and will not be changed without very good reasons.
-
-=over
-
-=item _get_cache
-
-This method will return the arguments that would have been determined from parsing the PATH_INFO
-(and possibly using the Dispatch Table), but that were stored in the cache from a previous run.
-It receives the URL of the request to use as the key:
-
-    my $args = $self->_get_cache($url)
-
-=item _set_cache
-
-This method sets the arguments determined form parsing the PATH_INFO to be used for this URL
-for future requests to the same URL.
-It receives the URL of the request (to be used as the key to the cache) and a reference to
-the data being stored:
-
-    $self->_set_cache($url, $args);
-
-=back
+#=head2 PROTECTED METHODS
+#
+#The following methods are intended to be overridden by subclasses if necessary. They are not
+#part of the public API since end users will never touch them. However, to ensure that your
+#subclass of Dispatch does not break with a new release, they are documented here and are considered
+#to be part of the API and will not be changed without very good reasons.
 
 =head1 AUTHOR
 
